@@ -3,12 +3,14 @@ package org.gcube.portal.databook.server;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Calendar;
+import java.util.Collections;
 import java.util.Date;
 import java.util.HashMap;
 import java.util.HashSet;
 import java.util.List;
 import java.util.Map;
 import java.util.Set;
+import java.util.concurrent.atomic.AtomicInteger;
 
 import javax.mail.internet.AddressException;
 import javax.mail.internet.InternetAddress;
@@ -42,6 +44,7 @@ import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
 import com.netflix.astyanax.MutationBatch;
+import com.netflix.astyanax.RowCallback;
 import com.netflix.astyanax.connectionpool.OperationResult;
 import com.netflix.astyanax.connectionpool.exceptions.ConnectionException;
 import com.netflix.astyanax.model.Column;
@@ -54,9 +57,8 @@ import com.netflix.astyanax.serializers.StringSerializer;
 import com.netflix.astyanax.util.RangeBuilder;
 /**
  * @author Massimiliano Assante ISTI-CNR
- * 
+ * @author Costantino Perciante ISTI-CNR
  * This class is used for querying and adding data to Cassandra via Astyanax High Level API 
- *
  */
 public final class DBCassandraAstyanaxImpl implements DatabookStore {
 
@@ -78,8 +80,8 @@ public final class DBCassandraAstyanaxImpl implements DatabookStore {
 	public static final String USER_TIMELINE_FEEDS = "USERTimeline";
 	public static final String APP_TIMELINE_FEEDS = "AppTimeline";
 	public static final String USER_LIKED_FEEDS = "USERLikes";
-	public static final String USER_NOTIFICATIONS = "USERNotifications"; // regular user notifications timeline
-	public static final String USER_MESSAGES_NOTIFICATIONS = "USERMessagesNotifications";  // user messages notifications timeline
+	public static final String USER_NOTIFICATIONS = "USERNotifications"; // regular user notifications timeline (both read and unread, messages are included)
+	public static final String USER_NOTIFICATIONS_UNREAD = "USERNotificationsUnread"; // only unread user notifications/ notifications messages
 	public static final String USER_NOTIFICATIONS_PREFERENCES = "USERNotificationsPreferences"; // preferences for notifications
 	public static final String HASHTAGS_COUNTER = "HashtagsCounter"; // count the hashtags per group and type
 	public static final String HASHTAGGED_FEEDS = "HashtaggedFeeds"; // contains hashtags per type associated with vre and feed
@@ -140,8 +142,8 @@ public final class DBCassandraAstyanaxImpl implements DatabookStore {
 			USER_NOTIFICATIONS,              // Column Family Name
 			StringSerializer.get(),   // Key Serializer
 			StringSerializer.get());  // Column Serializer
-	private static ColumnFamily<String, String> cf_UserMessageNotifications =	new ColumnFamily<String, String>(
-			USER_MESSAGES_NOTIFICATIONS,              // Column Family Name
+	private static ColumnFamily<String, String> cf_UserNotificationsUnread =	new ColumnFamily<String, String>(
+			USER_NOTIFICATIONS_UNREAD,              // Column Family Name
 			StringSerializer.get(),   // Key Serializer
 			StringSerializer.get());  // Column Serializer
 	protected static ColumnFamily<String, String> cf_UserNotificationsPreferences =	new ColumnFamily<String, String>(
@@ -192,6 +194,14 @@ public final class DBCassandraAstyanaxImpl implements DatabookStore {
 	public DBCassandraAstyanaxImpl() {
 		conn = new CassandraClusterConnection(false);
 	}
+
+	/**
+	 * public constructor, no dropping schema is allowed, infrastructureName is given.
+	 */
+	public DBCassandraAstyanaxImpl(String infrastructureName) {
+		conn = new CassandraClusterConnection(false, infrastructureName);
+	}
+
 	/**
 	 * execute the mutation batch
 	 * @param m
@@ -568,6 +578,44 @@ public final class DBCassandraAstyanaxImpl implements DatabookStore {
 	public List<Feed> getAllFeedsByApp(String appid) throws PrivacyLevelTypeNotFoundException, FeedTypeNotFoundException, ColumnNameNotFoundException, FeedIDNotFoundException {
 		return getFeedsByIds(getAppFeedIds(appid));
 	}
+	/**
+	 * {@inheritDoc}
+	 * @throws Exception 
+	 */
+	@Override
+	public List<Feed> getRecentCommentedFeedsByUserAndDate(String userid,
+			long timeInMillis) throws Exception {
+
+		List<Feed> toReturn = new ArrayList<Feed>();
+
+		Date now = new Date();
+		if (timeInMillis > now.getTime())
+			throw new IllegalArgumentException("the timeInMillis must be before today");
+
+		if(userid == null || userid.isEmpty())
+			throw new IllegalArgumentException("the userId parameter cannot be null/empty");
+
+		// get the last comments by the user (it is not needed to get them sorted)
+		List<Comment> lastComments = getRecentCommentsByUserAndDateBody(userid, timeInMillis, false);
+
+		// evaluate unique feeds' ids
+		HashSet<String> feedsIds = new HashSet<String>();
+
+		for (Comment comment : lastComments) {
+			String feedId = comment.getFeedid();
+			try{
+				if(!feedsIds.contains(feedId)){
+					feedsIds.add(feedId);
+					toReturn.add(readFeed(feedId));
+				}
+			}catch(Exception e){
+				_log.error("Unable to retrieve feed with id " + feedId, e);
+			}
+		}
+
+		Collections.sort(toReturn, Collections.reverseOrder());
+		return toReturn;
+	}
 	/**
 	 * helper method that retrieve all the feeds belongin to a list of Ids
 	 * @param feedIds
@@ -883,12 +931,11 @@ public final class DBCassandraAstyanaxImpl implements DatabookStore {
 		.putColumn("SenderFullName", n.getSenderFullName(), null)
 		.putColumn("SenderThumbnail", n.getSenderThumbnail(), null);
 
-		if (n.getType() != NotificationType.MESSAGE)
-			//an entry in the user Notifications Timeline 
-			m.withRow(cf_UserNotifications, n.getUserid()).putColumn(n.getTime().getTime()+"", n.getKey().toString(), null);
-		else
-			//an entry in the user Messages Notifications Timeline 
-			m.withRow(cf_UserMessageNotifications, n.getUserid()).putColumn(n.getTime().getTime()+"", n.getKey().toString(), null);
+		//an entry in the user Notifications Timeline 
+		m.withRow(cf_UserNotifications, n.getUserid()).putColumn(n.getTime().getTime()+"", n.getKey().toString(), null);
+
+		// save key in the unread notifications column family too
+		m.withRow(cf_UserNotificationsUnread, n.getUserid()).putColumn(n.getTime().getTime()+"", n.getKey().toString(), null);
 
 		return execute(m);
 	}
@@ -939,8 +986,14 @@ public final class DBCassandraAstyanaxImpl implements DatabookStore {
 			throw new NotificationIDNotFoundException("The specified notification to set Read with id: " + notificationidToSet + " does not exist");
 
 		MutationBatch m = conn.getKeyspace().prepareMutationBatch();
-		//an entry in the feed CF
+
+		//set as read
 		m.withRow(cf_Notifications, notificationidToSet).putColumn("Read", true, null);
+
+		// delete the notification's key from the unread notifications column family
+		m.withRow(cf_UserNotificationsUnread, toSet.getUserid()).deleteColumn(toSet.getTime().getTime()+"");
+
+		// execute the operations
 		try {
 			m.execute();
 		} catch (ConnectionException e) {
@@ -976,14 +1029,14 @@ public final class DBCassandraAstyanaxImpl implements DatabookStore {
 		return toReturn;
 	}
 	/**
-	 * 
+	 * Return a list of not read notifications by user userid (messages as well as other notifications)
 	 * @param userid  user identifier
-	 * @return simply return a list of user messages notifications UUID in chronological order from the oldest to the more recent
+	 * @return simply return a list of not read user notifications UUID in chronological order from the oldest to the more recent
 	 */
-	private ArrayList<String> getUserMessagesNotificationsIds(String userid) {
+	private ArrayList<String> getUnreadUserNotificationsIds(String userid) {
 		OperationResult<Rows<String, String>> result = null;
 		try {
-			result = conn.getKeyspace().prepareQuery(cf_UserMessageNotifications)
+			result = conn.getKeyspace().prepareQuery(cf_UserNotificationsUnread)
 					.getKeySlice(userid)
 					.execute();
 		} catch (ConnectionException e) {
@@ -1028,16 +1081,17 @@ public final class DBCassandraAstyanaxImpl implements DatabookStore {
 	@Override
 	public List<Notification> getUnreadNotificationsByUser(String userid) throws NotificationTypeNotFoundException,	ColumnNameNotFoundException, NotificationIDNotFoundException {
 		ArrayList<Notification> toReturn = new ArrayList<Notification>();
-		ArrayList<String> notificationsIDs = getUserNotificationsIds(userid);
+		ArrayList<String> notificationsIDs = getUnreadUserNotificationsIds(userid);
 
 		//need them in reverse order		
 		for (int i = notificationsIDs.size()-1; i >= 0; i--) {
-			Notification toAdd = readNotification(notificationsIDs.get(i));
-			if ((!toAdd.isRead()) && (toAdd.getType() != NotificationType.MESSAGE) ) { //i just set the first notification unread to read (much faster cuz i check only the newest)
-				toReturn.add(toAdd);
-				break;
+			try{
+				toReturn.add(readNotification(notificationsIDs.get(i)));
+			}catch(Exception e){
+				_log.error("Unable to read notification with key " + notificationsIDs.get(i));
 			}
 		}
+
 		return toReturn;	
 	}
 	/**
@@ -1077,64 +1131,42 @@ public final class DBCassandraAstyanaxImpl implements DatabookStore {
 	 */
 	@Override
 	public boolean setAllNotificationReadByUser(String userid) throws NotificationTypeNotFoundException, ColumnNameNotFoundException {
-		ArrayList<String> notificationsIDs = getUserNotificationsIds(userid);
 
-		//need them in reverse order		
+		// get the list of unread notifications
+		ArrayList<String> notificationsIDs = getUnreadUserNotificationsIds(userid);
+
 		for (int i = notificationsIDs.size()-1; i >= 0; i--) {
-			Notification toAdd;
-			try {
-				toAdd = readNotification(notificationsIDs.get(i));
-				if ((!toAdd.isRead()) && (toAdd.getType() != NotificationType.MESSAGE) ) {  //while I encounter unread notifications keep putting them to read, else exit
-					setNotificationRead(toAdd.getKey());
-				}
-				else {
-					break;
-				}
+			try{
+
+				// set to read (and automatically remove from the unread column family)
+				setNotificationRead(notificationsIDs.get(i));
+
 			} catch (NotificationIDNotFoundException e) {
 				_log.error("Could not set read notification with id =" + notificationsIDs.get(i));
 			}			
 		}
 		return true;	
 	}
-	/**
-	 * {@inheritDoc}
-	 */
-	@Override
-	public List<Notification> getUnreadNotificationMessagesByUser(String userid) throws NotificationTypeNotFoundException,	ColumnNameNotFoundException, NotificationIDNotFoundException {
-		ArrayList<Notification> toReturn = new ArrayList<Notification>();
-		ArrayList<String> notificationsIDs = getUserMessagesNotificationsIds(userid);
 
-		//need them in reverse order		
-		for (int i = notificationsIDs.size()-1; i >= 0; i--) {
-			Notification toAdd = readNotification(notificationsIDs.get(i));
-			if ((!toAdd.isRead()) && (toAdd.getType() == NotificationType.MESSAGE) ) {//i just set the first message notification unread to read (much faster)
-				toReturn.add(toAdd);
-				break;
-			}
-		}
-		return toReturn;	
-	}
 	/**
 	 * {@inheritDoc}
 	 */
 	@Override
 	public boolean checkUnreadNotifications(String userid) throws NotificationTypeNotFoundException, ColumnNameNotFoundException {	
-		ArrayList<String> notificationsIDs = getUserNotificationsIds(userid);
-		//since #readNotification costs time and newer notifications are iterarate first (with the reverse for below)
-		//i just see if the first non message notification (UserNotifications TimeLine) is read or not and return the value instead of iterating them one by one looking for unread()
 
-		//need them in reverse order		
-		for (int i = notificationsIDs.size()-1; i >= 0; i--) {
+		ArrayList<String> unreadNotifications = getUnreadUserNotificationsIds(userid);
+
+		for (int i = unreadNotifications.size() - 1; i >= 0; i--) {
 			Notification toAdd;
 			try {
-				toAdd = readNotification(notificationsIDs.get(i));
+				toAdd = readNotification(unreadNotifications.get(i));
 				if (toAdd.getType() != NotificationType.MESSAGE)
-					return ! toAdd.isRead();
+					return true;
 			} catch (NotificationIDNotFoundException e) {
-				_log.error("Notification not found with id = " + notificationsIDs.get(i));
-				return false;
-			}			
+				_log.error("Notification not found with id = " + unreadNotifications.get(i));
+			}
 		}
+
 		return false;
 	}
 	/**
@@ -1142,16 +1174,20 @@ public final class DBCassandraAstyanaxImpl implements DatabookStore {
 	 */
 	@Override
 	public boolean checkUnreadMessagesNotifications(String userid) throws NotificationIDNotFoundException, NotificationTypeNotFoundException, ColumnNameNotFoundException {
-		ArrayList<String> notificationsIDs = getUserMessagesNotificationsIds(userid);
-		//since #readNotification costs time and newer notifications are iterarate first (with the reverse for below)
-		//i just see if the first message notification (UserMessagesNotifications TL) is read or not and return the value instead of iterating them one by one looking for unread()
 
-		//need them in reverse order		
-		for (int i = notificationsIDs.size()-1; i >= 0; i--) {
-			Notification toAdd = readNotification(notificationsIDs.get(i));
-			if (toAdd.getType() == NotificationType.MESSAGE)
-				return ! toAdd.isRead();
+		ArrayList<String> unreadNotifications = getUnreadUserNotificationsIds(userid);
+
+		for (int i = unreadNotifications.size() - 1; i >= 0; i--) {
+			Notification toAdd;
+			try {
+				toAdd = readNotification(unreadNotifications.get(i));
+				if (toAdd.getType() == NotificationType.MESSAGE)
+					return true;
+			} catch (NotificationIDNotFoundException e) {
+				_log.error("Notification not found with id = " + unreadNotifications.get(i));
+			}
 		}
+
 		return false;
 	}
 	/*
@@ -1419,6 +1455,95 @@ public final class DBCassandraAstyanaxImpl implements DatabookStore {
 		return toReturn;
 	}
 
+	/**
+	 * {@inheritDoc}
+	 * @throws Exception 
+	 */
+	@Override
+	public List<Comment> getRecentCommentsByUserAndDate(final String userid,
+			final long timeInMillis) throws Exception {
+
+		final List<Comment> commentsByUser;
+
+		Date now = new Date();
+		if (timeInMillis > now.getTime())
+			throw new IllegalArgumentException("the timeInMillis must be before today");
+
+		if(userid == null || userid.isEmpty())
+			throw new IllegalArgumentException("the userId parameter cannot be null/empty");
+
+
+		commentsByUser = getRecentCommentsByUserAndDateBody(userid, timeInMillis, true);
+
+		return commentsByUser;
+	}
+
+	/**
+	 * Private method that allows also to specify if the returned list must be sorted or not
+	 * @param userid the user id
+	 * @param timeInMillis the initial time to consider
+	 * @param sort a boolean value to specify if the returned list must be sorted (from the most recent to the oldest comment)
+	 * @return a list of comments recently made by the user
+	 * @throws ConnectionException 
+	 */
+	private List<Comment> getRecentCommentsByUserAndDateBody(final String userid,
+			final long timeInMillis, boolean sort) throws ConnectionException{
+
+		final List<Comment> commentsByUser = new ArrayList<Comment>();
+		final AtomicInteger maxRetryExecutions = new AtomicInteger(10);
+
+		/**
+		 * See https://github.com/Netflix/astyanax/wiki/Reading-Data for more information (Query all with callback)
+		 */
+		conn.getKeyspace().prepareQuery(cf_Comments)
+		.getAllRows()
+		.setRowLimit(100)  // Read in blocks of 100
+		.withColumnRange(new RangeBuilder().setLimit(10).build())
+		.executeWithCallback(new RowCallback<String, String>() {
+			@Override
+			public void success(Rows<String, String> rows) {
+				synchronized (commentsByUser) {
+					nextRow: for (Row<String, String> row : rows) {
+						for(Column<String> column: row.getColumns())
+							if(column.getName().equals("Userid"))
+								if(column.getStringValue().equals(userid)){
+									try{
+										Comment c = readCommentById(row.getKey());
+										Feed f = readFeed(c.getFeedid());
+										if(c.getTime().getTime() >= timeInMillis && 
+												(f.getType() == FeedType.TWEET || f.getType() == FeedType.SHARE || f.getType() == FeedType.PUBLISH))
+											commentsByUser.add(c);
+									}catch(Exception e){
+										_log.error("Unable to read comment with id" + row.getKey(), e);
+									}
+									continue nextRow;
+								}
+					}
+				}
+			}
+
+			@Override
+			public boolean failure(ConnectionException e) {
+				// decrement value
+				int currentValue = maxRetryExecutions.decrementAndGet();
+				if(currentValue <= 0){
+					_log.error("Error while fetching user's recent comments ... repeating operation");
+					return true;  
+				}else
+				{
+					_log.error("Too many errors while fetching user's comments, exiting");
+					return false; 
+				}
+			}
+		});
+
+		if(sort)
+			Collections.sort(commentsByUser, Collections.reverseOrder());
+
+		return commentsByUser;
+
+	}
+
 	/**
 	 * {@inheritDoc}
 	 */
@@ -1590,6 +1715,57 @@ public final class DBCassandraAstyanaxImpl implements DatabookStore {
 		return toReturn;
 	}
 
+	/**
+	 * {@inheritDoc}
+	 */
+	@Override
+	public List<Feed> getRecentLikedFeedsByUserAndDate(String userid,
+			long timeInMillis) throws IllegalArgumentException {
+
+		List<Feed> toReturn = new ArrayList<Feed>();
+
+		Date now = new Date();
+		if (timeInMillis > now.getTime())
+			throw new IllegalArgumentException("the timeInMillis must be before today");
+
+		if(userid == null || userid.isEmpty())
+			throw new IllegalArgumentException("the userId parameter cannot be null/empty");
+
+		// get the list of liked feeds
+		List<String> likedFeedsIdsByUser = getAllLikedFeedIdsByUser(userid);
+
+		if(likedFeedsIdsByUser != null){
+
+			for(int i = likedFeedsIdsByUser.size() - 1; i >= 0; i--){
+				String feedId = likedFeedsIdsByUser.get(i);
+				try{
+
+					// retrieve the feed
+					Feed toCheck = readFeed(feedId);
+					boolean isFeedOk = (toCheck.getType() == FeedType.TWEET || toCheck.getType() == FeedType.SHARE || toCheck.getType() == FeedType.PUBLISH);
+
+					// retrieve the like of the user for the feed
+					if(isFeedOk){
+						List<Like> likes = getAllLikesByFeed(feedId);
+						for (Like like : likes) {
+							if(like.getTime().getTime() >= timeInMillis && like.getUserid().equals(userid))
+								toReturn.add(toCheck);
+						}
+					}
+
+				}catch(Exception e){
+					_log.error("Skipped feed with id " + feedId, e);
+				}
+			}
+		}
+
+		// please check consider that if a user made like recently to an old post, well it could happen that this 
+		// post comes first than a newer post in the toReturn list. Thus we need to sort it.
+		Collections.sort(toReturn, Collections.reverseOrder());
+
+		return toReturn;
+	}
+
 	/**
 	 * {@inheritDoc}
 	 */
@@ -1707,6 +1883,59 @@ public final class DBCassandraAstyanaxImpl implements DatabookStore {
 		}
 		return toReturn;
 	}
+	/**
+	 * {@inheritDoc}
+	 */
+	@Override
+	public Map<String, Integer> getVREHashtagsWithOccurrenceFilteredByTime(String vreid, long timestamp){
+		OperationResult<Rows<String, String>> result = null;
+		try {
+			result = conn.getKeyspace().prepareQuery(cf_HashtagsCounter)
+					.getKeySlice(vreid)
+					.execute();
+		} catch (ConnectionException e) {
+			e.printStackTrace();
+		}
+
+		HashMap<String, Integer> toReturn = new HashMap<String, Integer> ();
+
+		// Iterate rows and their columns 
+		for (Row<String, String> row : result.getResult()) {
+			for (Column<String> column : row.getColumns()) {
+
+				// retrieve the feeds list for this hashtag
+				List<Feed> feeds = null;
+				try{
+					feeds = getVREFeedsByHashtag(vreid, column.getName());
+				}catch(Exception e){
+					_log.error("Unable to retrieve the list of feeds for hashtag" + column.getName() + " in vre " + vreid);
+					continue;
+				}
+
+				if(feeds.isEmpty()){
+
+					_log.info("There are no feeds containing hashtag " + column.getName() + " in  vre " + vreid);
+					continue;
+
+				}
+
+				// retrieve the most recent one among these feeds
+				Collections.sort(feeds, Collections.reverseOrder());
+
+				if(feeds.get(0).getTime().getTime() < timestamp){
+					continue;
+				}
+
+				// else..
+				int curValue = Integer.parseInt(column.getStringValue());
+
+				if (curValue > 0)
+					toReturn.put(column.getName(), curValue);
+			}
+		}
+		return toReturn;
+	}
+
 	/**
 	 * {@inheritDoc}
 	 */
@@ -1972,6 +2201,7 @@ public final class DBCassandraAstyanaxImpl implements DatabookStore {
 	 */
 	@Override
 	public void closeConnection() {
+		conn.closeConnection();
 	}
 	/*
 	 * 
@@ -2358,5 +2588,4 @@ public final class DBCassandraAstyanaxImpl implements DatabookStore {
 		return ids;
 
 	}
-
 }
